#!/bin/bash
set -e
set -o pipefail
set -x

machine=$2
cortex=$3
iter=3

echoerr() { echo "$@" 1>&2; }

function bench_eb {
	cmake -DFB_POLYN=$1 -DEB_WIDTH=$2 -DEB_DEPTH=$3 -DMD_METHD=$4 -DEC_KBLTZ=ON -DEB_KBLTZ=ON \
	"-DEB_METHD=PROJC;LODAH;COMBS;INTER" \
	"-DCOMP=-O2 -g -mcpu=cortex-a$cortex -march=armv7-a -mfpu=neon -mfloat-abi=softfp -marm -Deb_choose=eb_choose_asm -DCORTEX=$cortex" . &> conf$1.log
	echoerr Compiling $1
	make bench_fb bench_eb &> make$1.log
	echoerr Running fb $1
	for ((i = 1; i <= $iter; i++)); do
		maketesta.sh bench_fb $machine cs &> runfb$1.log
		grep -E 'BENCH|z\^' runfb$1.log
	done
	echoerr Running eb $1
	for ((i = 1; i <= $iter; i++)); do
		maketesta.sh bench_eb $machine cs &> runeb$1.log
		grep -E 'BENCH|-- Curve' runeb$1.log
	done
#	if [ $5 == "ON" ]; then
#		cmake "-DEB_METHD=PROJC;LWNAF;COMBS;INTER" . &>> conf$1.log
#		echoerr Running cp $1
#		for ((i = 1; i <= $iter; i++)); do
#			maketesta.sh bench_cp $machine cs &> runcp$1.log
#			grep -E 'BENCH|-- Curve' runcp$1.log
#		done
#	fi
#	cmake -DEC_KBLTZ=OFF -DEB_KBLTZ=OFF "-DEB_METHD=PROJC;LODAH;COMBS;INTER" . &>> conf$1.log
	echoerr Running cp $1
	for ((i = 1; i <= $iter; i++)); do
		maketesta.sh bench_cp $machine cs &> runcp$1.log
		grep -E 'BENCH|-- Curve' runcp$1.log
	done
}

function bench_pb {
	cmake -DFB_POLYN=$1 -DEB_WIDTH=$2 -DEB_DEPTH=$3 -DMD_METHD=$4 \
	"-DEB_METHD=PROJC;LWNAF;COMBS;INTER" \
	"-DCOMP=-O2 -g -mcpu=cortex-a$cortex -march=armv7-a -mfpu=neon -mfloat-abi=softfp -marm -Deb_choose=eb_choose_asm -DCORTEX=$cortex" . &> conf$1.log
	echoerr Compiling $1
	make bench_fb bench_pb &> make$1.log
	echoerr Running fb $1
	for ((i = 1; i <= $iter; i++)); do
		maketesta.sh bench_fb $machine cs &> runfb$1.log
		grep -E 'BENCH|z\^' runfb$1.log
	done
	echoerr Running pb $1
	for ((i = 1; i <= $iter; i++)); do
		maketesta.sh bench_pb $machine cs &> runpb$1.log
		grep -E 'BENCH|-- Curve' runpb$1.log
	done
}

$1/preset/arm-ecc-128.sh $1 &> conf.log

bench_eb 251 5 8 SH256 OFF
#bench_pb 353 5 8 SH256
bench_eb 163 5 8 SH256 ON
bench_eb 571 5 8 SH512 ON
bench_eb 283 5 8 SH256 ON
