#ifndef ECC_GF_H_
#define ECC_GF_H_

#define MP6_T64 3
#define MP8_T64 4
#define MP9_T64 5
#define MP12_T64 6
#define MP13_T64 7
#define MP18_T64 9

typedef unsigned mp6_t[MP6_T64*2] __attribute__ ((aligned (8)));
typedef unsigned mp8_t[MP8_T64*2] __attribute__ ((aligned (8)));
typedef unsigned mp9_t[MP9_T64*2] __attribute__ ((aligned (8)));
typedef unsigned mp12_t[MP12_T64*2] __attribute__ ((aligned (8)));
typedef unsigned mp13_t[MP13_T64*2] __attribute__ ((aligned (8)));
typedef unsigned mp18_t[MP18_T64*2] __attribute__ ((aligned (8)));

typedef struct {
    mp6_t x;
    mp6_t y;
} mp6_2t;

typedef struct {
    mp8_t x;
    mp8_t y;
} mp8_2t;

typedef struct {
    mp9_t x;
    mp9_t y;
} mp9_2t;

typedef struct {
    mp12_t x;
    mp12_t y;
} mp12_2t;

typedef struct {
    mp13_t x;
    mp13_t y;
} mp13_2t;

typedef struct {
    mp18_t x;
    mp18_t y;
} mp18_2t;

void mp6_eor  (mp6_t r, const mp6_t a, const mp6_t b);
void mp8_eor  (mp8_t r, const mp8_t a, const mp8_t b);
void mp9_eor  (mp9_t r, const mp9_t a, const mp9_t b);
void mp12_eor (mp12_t r, const mp12_t a, const mp12_t b);
void mp13_eor (mp13_t r, const mp13_t a, const mp13_t b);
void mp18_eor (mp18_t r, const mp18_t a, const mp18_t b);

unsigned mp6_lsr_1  (mp6_t r, const mp6_t a);
unsigned mp6_lsl_1  (mp6_t r, const mp6_t a);
unsigned mp6_lsr_32 (mp6_t r, const mp6_t a, unsigned u);
unsigned mp6_lsl_32 (mp6_t r, const mp6_t a, unsigned u);

unsigned mp8_lsr_1  (mp8_t r, const mp8_t a);
unsigned mp8_lsl_1  (mp8_t r, const mp8_t a);
unsigned mp8_lsr_32 (mp8_t r, const mp8_t a, unsigned u);
unsigned mp8_lsl_32 (mp8_t r, const mp8_t a, unsigned u);

unsigned mp9_lsr_1  (mp9_t r, const mp9_t a);
unsigned mp9_lsl_1  (mp9_t r, const mp9_t a);
unsigned mp9_lsr_32 (mp9_t r, const mp9_t a, unsigned u);
unsigned mp9_lsl_32 (mp9_t r, const mp9_t a, unsigned u);

unsigned mp12_lsr_1  (mp12_t r, const mp12_t a);
unsigned mp12_lsl_1  (mp12_t r, const mp12_t a);
unsigned mp12_lsr_32 (mp12_t r, const mp12_t a, unsigned u);
unsigned mp12_lsl_32 (mp12_t r, const mp12_t a, unsigned u);

unsigned mp13_lsr_1  (mp13_t r, const mp13_t a);
unsigned mp13_lsl_1  (mp13_t r, const mp13_t a);
unsigned mp13_lsr_32 (mp13_t r, const mp13_t a, unsigned u);
unsigned mp13_lsl_32 (mp13_t r, const mp13_t a, unsigned u);

unsigned mp18_lsr_1  (mp18_t r, const mp18_t a);
unsigned mp18_lsl_1  (mp18_t r, const mp18_t a);
unsigned mp18_lsr_32 (mp18_t r, const mp18_t a, unsigned u);
unsigned mp18_lsl_32 (mp18_t r, const mp18_t a, unsigned u);


//------------------------------------------------------------------------------
// NIST B163
// p(z) = z^{163} + z^7 + z^6 + z^3 + 1

static mp6_t gf_b163_b = {0x4a3205fd, 0x512f7874, 0x1481eb10, 0xb8c953ca, 0x0a601907, 0x2};

#define gf_b163_add mp6_eor

void gf_b163_mod (mp6_t r, unsigned* a);

void gf_b163_square_t (mp6_t r, const mp6_t a);
void gf_b163_square_v (mp6_t r, const mp6_t a);
#define gf_b163_square gf_b163_square_v

void gf_b163_mul_kld1 (mp6_t r, const mp6_t a, const mp6_t b);
void gf_b163_mul_kld  (mp6_t r, const mp6_t a, const mp6_t b);
void gf_b163_mul_kldu (mp6_t r, const mp6_t a, const mp6_t b);
void gf_b163_mul_ld   (mp6_t r, const mp6_t a, const mp6_t b);
void gf_b163_mul_ldu  (mp6_t r, const mp6_t a, const mp6_t b);
void gf_b163_mul_vld  (mp6_t r, const mp6_t a, const mp6_t b);
void gf_b163_mul_v    (mp6_t r, const mp6_t a, const mp6_t b);
void gf_b163_muld_vld (unsigned* r, const mp6_t a, const mp6_t b);
void gf_b163_muld_v   (unsigned* r, const mp6_t a, const mp6_t b);
#define gf_b163_mul   gf_b163_mul_v
#define gf_b163_muld   gf_b163_muld_v
#define gf_b163_mul_b(r, a) gf_b163_mul(r, a, gf_b163_b)

void gf_b163_inv_eeb (mp6_t r, const mp6_t a);
void gf_b163_inv_ee  (mp6_t r, const mp6_t a);
void gf_b163_inv_vee (mp6_t r, const mp6_t a);
#define gf_b163_inv  gf_b163_inv_vee

void gf_b163_sqrt   (mp6_t r, const mp6_t a);

void gf_b163_solve_b    (mp6_t r, const mp6_t a);
void gf_b163_solve_rm2  (mp6_t r, const mp6_t a);
void gf_b163_solve_4m2  (mp6_t r, const mp6_t a);
void gf_b163_solve_vb   (mp6_t r, const mp6_t a);
void gf_b163_solve_vrm2 (mp6_t r, const mp6_t a);
void gf_b163_solve_v4m2 (mp6_t r, const mp6_t a);
#define gf_b163_solve   gf_b163_solve_v4m2

int gf_b163_degree (const mp6_t a, int size);

// Trace function of a field element: Tr(z^i) = 1 only for i \in {0, 157} -> Tr(a) = a_157 + a_0
#define gf_b163_trace(a) (((a[157>>5] >> (157%32)) ^ a[0]) & 1)


//------------------------------------------------------------------------------
// NIST B233
// p(z) = z^{233} + z^{74} + 1

static mp8_t gf_b233_b = {0x7d8f90ad, 0x81fe115f, 0x20e9ce42, 0x213b333b, 0x0923bb58, 0x332c7f8c, 0x647ede6c, 0x066};

#define gf_b233_add mp8_eor

void gf_b233_mod (mp8_t r, unsigned* a);

void gf_b233_square_t (mp8_t r, const mp8_t a);
void gf_b233_square_v (mp8_t r, const mp8_t a);
#define gf_b233_square gf_b233_square_v

void gf_b233_mul_kld1 (mp8_t r, const mp8_t a, const mp8_t b);
void gf_b233_mul_kld  (mp8_t r, const mp8_t a, const mp8_t b);
void gf_b233_mul_kldu (mp8_t r, const mp8_t a, const mp8_t b);
void gf_b233_mul_ld   (mp8_t r, const mp8_t a, const mp8_t b);
void gf_b233_mul_ldu  (mp8_t r, const mp8_t a, const mp8_t b);
void gf_b233_mul_vld  (mp8_t r, const mp8_t a, const mp8_t b);
void gf_b233_mul_v    (mp8_t r, const mp8_t a, const mp8_t b);
void gf_b233_muld_vld (unsigned* r, const mp8_t a, const mp8_t b);
void gf_b233_muld_v   (unsigned* r, const mp8_t a, const mp8_t b);
#define gf_b233_mul   gf_b233_mul_v
#define gf_b233_muld   gf_b233_muld_v
#define gf_b233_mul_b(r, a) gf_b233_mul(r, a, gf_b233_b)

void gf_b233_inv_eeb (mp8_t r, const mp8_t a);
void gf_b233_inv_ee  (mp8_t r, const mp8_t a);
void gf_b233_inv_vee (mp8_t r, const mp8_t a);
#define gf_b233_inv  gf_b233_inv_ee

void gf_b233_sqrt (mp8_t r, const mp8_t a);

void gf_b233_solve_b    (mp8_t r, const mp8_t a);
void gf_b233_solve_rm2  (mp8_t r, const mp8_t a);
void gf_b233_solve_4m2  (mp8_t r, const mp8_t a);
void gf_b233_solve_vb   (mp8_t r, const mp8_t a);
void gf_b233_solve_vrm2 (mp8_t r, const mp8_t a);
void gf_b233_solve_v4m2 (mp8_t r, const mp8_t a);
#define gf_b233_solve   gf_b233_solve_v4m2

int gf_b233_degree (const mp8_t a, int size);

// Trace function of a field element: Tr(z^i) = 1 only for i \in {0, 159} -> Tr(a) = a_159 + a_0
#define gf_b233_trace(a) (((a[159>>5] >> (159%32)) ^ a[0]) & 1)


//------------------------------------------------------------------------------
// NIST B283
// p(z) = z^{283} + z^{12} + z^7 + z^5 + 1

static mp9_t gf_b283_b = {0x3b79a2f5, 0xf6263e31, 0xa581485a, 0x45309fa2, 0xca97fd76, 0x19a0303f, 0xa5a4af8a, 0xc8b8596d, 0x027b680a};

#define gf_b283_add mp9_eor

void gf_b283_mod (mp9_t r, unsigned* a);

void gf_b283_square_t (mp9_t r, const mp9_t a);
void gf_b283_square_v (mp9_t r, const mp9_t a);
#define gf_b283_square gf_b283_square_v

void gf_b283_mul_kld1 (mp9_t r, const mp9_t a, const mp9_t b);
void gf_b283_mul_kld  (mp9_t r, const mp9_t a, const mp9_t b);
void gf_b283_mul_kldu (mp9_t r, const mp9_t a, const mp9_t b);
void gf_b283_mul_ld   (mp9_t r, const mp9_t a, const mp9_t b);
void gf_b283_mul_ldu  (mp9_t r, const mp9_t a, const mp9_t b);
void gf_b283_mul_vld  (mp9_t r, const mp9_t a, const mp9_t b);
void gf_b283_mul_v    (mp9_t r, const mp9_t a, const mp9_t b);
void gf_b283_muld_vld (unsigned* r, const mp9_t a, const mp9_t b);
void gf_b283_muld_v   (unsigned* r, const mp9_t a, const mp9_t b);
#define gf_b283_mul   gf_b283_mul_v
#define gf_b283_muld   gf_b283_muld_v
#define gf_b283_mul_b(r, a) gf_b283_mul(r, a, gf_b283_b)

void gf_b283_inv_eeb (mp9_t r, const mp9_t a);
void gf_b283_inv_ee  (mp9_t r, const mp9_t a);
void gf_b283_inv_vee (mp9_t r, const mp9_t a);
#define gf_b283_inv  gf_b283_inv_vee

void gf_b283_sqrt (mp9_t r, const mp9_t a);

void gf_b283_solve_b    (mp9_t r, const mp9_t a);
void gf_b283_solve_rm2  (mp9_t r, const mp9_t a);
void gf_b283_solve_4m2  (mp9_t r, const mp9_t a);
void gf_b283_solve_vb   (mp9_t r, const mp9_t a);
void gf_b283_solve_vrm2 (mp9_t r, const mp9_t a);
void gf_b283_solve_v4m2 (mp9_t r, const mp9_t a);
#define gf_b283_solve   gf_b283_solve_v4m2

int gf_b283_degree (const mp9_t a, int size);

// Trace function of a field element: Tr(z^i) = 1 only for i \in {0, 271} -> Tr(a) = a_271 + a_0
#define gf_b283_trace(a) (((a[271>>5] >> (271%32)) ^ a[0]) & 1)


//------------------------------------------------------------------------------
// NIST B409
// p(z) = z^{409} + z^{87} + 1

static mp13_t gf_b409_b = {0x7b13545f, 0x4f50ae31, 0xd57a55aa, 0x72822f6c, 0xa9a197b2, 0xd6ac27c8, 0x4761fa99, 0xf1f3dd67, 0x7fd6422e, 0x3b7b476b, 0x5c4b9a75, 0xc8ee9feb, 0x0021a5c2};

#define gf_b409_add mp13_eor

void gf_b409_mod (mp13_t r, unsigned* a);

void gf_b409_square_t (mp13_t r, const mp13_t a);
void gf_b409_square_v (mp13_t r, const mp13_t a);
#define gf_b409_square gf_b409_square_v

void gf_b409_mul_kld1 (mp13_t r, const mp13_t a, const mp13_t b);
void gf_b409_mul_kld  (mp13_t r, const mp13_t a, const mp13_t b);
void gf_b409_mul_kldu (mp13_t r, const mp13_t a, const mp13_t b);
void gf_b409_mul_ld   (mp13_t r, const mp13_t a, const mp13_t b);
void gf_b409_mul_ldu  (mp13_t r, const mp13_t a, const mp13_t b);
void gf_b409_mul_vld  (mp13_t r, const mp13_t a, const mp13_t b);
void gf_b409_mul_v    (mp13_t r, const mp13_t a, const mp13_t b);
void gf_b409_muld_vld (unsigned* r, const mp13_t a, const mp13_t b);
void gf_b409_muld_v   (unsigned* r, const mp13_t a, const mp13_t b);
#define gf_b409_mul   gf_b409_mul_v
#define gf_b409_muld   gf_b409_muld_v
#define gf_b409_mul_b(r, a) gf_b409_mul(r, a, gf_b409_b)

void gf_b409_inv_eeb (mp13_t r, const mp13_t a);
void gf_b409_inv_ee  (mp13_t r, const mp13_t a);
void gf_b409_inv_vee (mp13_t r, const mp13_t a);
#define gf_b409_inv  gf_b409_inv_vee

void gf_b409_sqrt (mp13_t r, const mp13_t a);

void gf_b409_solve_b    (mp13_t r, const mp13_t a);
void gf_b409_solve_rm2  (mp13_t r, const mp13_t a);
void gf_b409_solve_4m2  (mp13_t r, const mp13_t a);
void gf_b409_solve_vb   (mp13_t r, const mp13_t a);
void gf_b409_solve_vrm2 (mp13_t r, const mp13_t a);
void gf_b409_solve_v4m2 (mp13_t r, const mp13_t a);
#define gf_b409_solve   gf_b409_solve_v4m2

int gf_b409_degree (const mp13_t a, int size);

// Trace function of a field element: Tr(z^i) = 1 only for i \in {0} -> Tr(a) = a_0
#define gf_b409_trace(a) (a[0] & 1)


//------------------------------------------------------------------------------
// NIST B571
// p(z) = z^{571} + z^{10} + z^5 + z^2 + 1

static mp18_t gf_b571_b = {0x2955727a, 0x7ffeff7f, 0x39baca0c, 0x520e4de7, 0x78ff12aa, 0x4afd185a, 0x56a66e29, 0x2be7ad67, 0x8efa5933, 0x84ffabbd, 0x4a9a18ad, 0xcd6ba8ce, 0xcb8ceff1, 0x5c6a97ff, 0xb7f3d62f, 0xde297117, 0x2221f295, 0x02f40e7e};

#define gf_b571_add mp18_eor

void gf_b571_mod (mp18_t r, unsigned* a);

void gf_b571_square_t (mp18_t r, const mp18_t a);
void gf_b571_square_v (mp18_t r, const mp18_t a);
#define gf_b571_square gf_b571_square_v

void gf_b571_mul_kld1 (mp18_t r, const mp18_t a, const mp18_t b);
void gf_b571_mul_kld  (mp18_t r, const mp18_t a, const mp18_t b);
void gf_b571_mul_kldu (mp18_t r, const mp18_t a, const mp18_t b);
void gf_b571_mul_ld   (mp18_t r, const mp18_t a, const mp18_t b);
void gf_b571_mul_ldu  (mp18_t r, const mp18_t a, const mp18_t b);
void gf_b571_mul_vld  (mp18_t r, const mp18_t a, const mp18_t b);
void gf_b571_mul_v    (mp18_t r, const mp18_t a, const mp18_t b);
void gf_b571_muld_vld (unsigned* r, const mp18_t a, const mp18_t b);
void gf_b571_muld_v   (unsigned* r, const mp18_t a, const mp18_t b);
#define gf_b571_mul   gf_b571_mul_v
#define gf_b571_muld   gf_b571_muld_v
#define gf_b571_mul_b(r, a) gf_b571_mul(r, a, gf_b571_b)

void gf_b571_inv_eeb (mp18_t r, const mp18_t a);
void gf_b571_inv_ee  (mp18_t r, const mp18_t a);
void gf_b571_inv_vee (mp18_t r, const mp18_t a);
#define gf_b571_inv  gf_b571_inv_vee

void gf_b571_sqrt (mp18_t r, const mp18_t a);

void gf_b571_solve_b    (mp18_t r, const mp18_t a);
void gf_b571_solve_rm2  (mp18_t r, const mp18_t a);
void gf_b571_solve_4m2  (mp18_t r, const mp18_t a);
void gf_b571_solve_vb   (mp18_t r, const mp18_t a);
void gf_b571_solve_vrm2 (mp18_t r, const mp18_t a);
void gf_b571_solve_v4m2 (mp18_t r, const mp18_t a);
#define gf_b571_solve   gf_b571_solve_vrm2

int gf_b571_degree (const mp18_t a, int size);

// Trace function of a field element: Tr(z^i) = 1 only for i \in {0, 561, 569} -> Tr(a) = a_569 + a_561 + a_0
#define gf_b571_trace(a) (((a[569>>5] >> (569%32)) ^ (a[561>>5] >> (561%32)) ^ a[0]) & 1)


//------------------------------------------------------------------------------
// eBATS B251
// p(z) = z^{251} + z^7 + z^4 + z^2 + 1

#define gf_r251_add mp8_eor

void gf_r251_mod (mp8_t r, unsigned* a);

void gf_r251_square_t (mp8_t r, const mp8_t a);
void gf_r251_square_v (mp8_t r, const mp8_t a);
#define gf_r251_square gf_r251_square_v

void gf_r251_mul_kld1 (mp8_t r, const mp8_t a, const mp8_t b);
void gf_r251_mul_kld  (mp8_t r, const mp8_t a, const mp8_t b);
void gf_r251_mul_kldu (mp8_t r, const mp8_t a, const mp8_t b);
void gf_r251_mul_ld   (mp8_t r, const mp8_t a, const mp8_t b);
void gf_r251_mul_ldu  (mp8_t r, const mp8_t a, const mp8_t b);
void gf_r251_mul_vld  (mp8_t r, const mp8_t a, const mp8_t b);
void gf_r251_mul_v    (mp8_t r, const mp8_t a, const mp8_t b);
void gf_r251_muld_vld (unsigned* r, const mp8_t a, const mp8_t b);
void gf_r251_muld_v   (unsigned* r, const mp8_t a, const mp8_t b);
#define gf_r251_mul   gf_r251_mul_v
#define gf_r251_muld   gf_r251_muld_v

void gf_r251_mul_b_c  (mp8_t r, const mp8_t a);
void gf_r251_mul_b_v  (mp8_t r, const mp8_t a);
#define gf_r251_mul_b gf_r251_mul_b_v

void gf_r251_inv_eeb (mp8_t r, const mp8_t a);
void gf_r251_inv_ee  (mp8_t r, const mp8_t a);
void gf_r251_inv_vee (mp8_t r, const mp8_t a);
#define gf_r251_inv  gf_r251_inv_vee

void gf_r251_sqrt (mp8_t r, const mp8_t a);

void gf_r251_solve_b    (mp8_t r, const mp8_t a);
void gf_r251_solve_rm2  (mp8_t r, const mp8_t a);
void gf_r251_solve_4m2  (mp8_t r, const mp8_t a);
void gf_r251_solve_vb   (mp8_t r, const mp8_t a);
void gf_r251_solve_vrm2 (mp8_t r, const mp8_t a);
void gf_r251_solve_v4m2 (mp8_t r, const mp8_t a);
#define gf_r251_solve   gf_r251_solve_v4m2

int gf_r251_degree (const mp8_t a, int size);

// Trace function of a field element: Tr(z^i) = 1 only for i \in {0, 247, 249} -> Tr(a) = a_249 + a_247 + a_0
#define gf_r251_trace(a) (((a[249>>5] >> (249%32)) ^ ((a[247>>5] >> (247%32)) ^ a[0])) & 1)


//------------------------------------------------------------------------------
// S271
// p(z) = z^{271} + z^{201} + 1

// FIXME dummy value
static mp9_t gf_s271_b = {0x3b79a2f5, 0xf6263e31, 0xa581485a, 0x45309fa2, 0xca97fd76, 0x19a0303f, 0xa5a4af8a, 0xc8b8596d, 0x0000680a};

#define gf_s271_add mp9_eor

void gf_s271_mod (mp9_t r, unsigned* a);

void gf_s271_square_t (mp9_t r, const mp9_t a);
void gf_s271_square_v (mp9_t r, const mp9_t a);
#define gf_s271_square gf_s271_square_v

void gf_s271_mul_kld1 (mp9_t r, const mp9_t a, const mp9_t b);
void gf_s271_mul_kld  (mp9_t r, const mp9_t a, const mp9_t b);
void gf_s271_mul_kldu (mp9_t r, const mp9_t a, const mp9_t b);
void gf_s271_mul_ld   (mp9_t r, const mp9_t a, const mp9_t b);
void gf_s271_mul_ldu  (mp9_t r, const mp9_t a, const mp9_t b);
void gf_s271_mul_vld  (mp9_t r, const mp9_t a, const mp9_t b);
void gf_s271_mul_v    (mp9_t r, const mp9_t a, const mp9_t b);
void gf_s271_muld_vld (unsigned* r, const mp9_t a, const mp9_t b);
void gf_s271_muld_v   (unsigned* r, const mp9_t a, const mp9_t b);
#define gf_s271_mul   gf_s271_mul_v
#define gf_s271_muld   gf_s271_muld_v
#define gf_s271_mul_b(r, a) gf_s271_mul(r, a, gf_s271_b)

void gf_s271_inv_eeb (mp9_t r, const mp9_t a);
void gf_s271_inv_ee  (mp9_t r, const mp9_t a);
void gf_s271_inv_vee (mp9_t r, const mp9_t a);
#define gf_s271_inv  gf_s271_inv_ee

void gf_s271_sqrt (mp9_t r, const mp9_t a);

void gf_s271_solve_b    (mp9_t r, const mp9_t a);
void gf_s271_solve_4m2  (mp9_t r, const mp9_t a);
void gf_s271_solve_vb   (mp9_t r, const mp9_t a);
void gf_s271_solve_v4m2 (mp9_t r, const mp9_t a);
#define gf_s271_solve   gf_s271_solve_v4m2

int gf_s271_degree (const mp9_t a, int size);

// Trace function of a field element: Tr(z^i) = 1 only for i \in {0} -> Tr(a) = a_0
#define gf_s271_trace(a) (a[0] & 1)


//------------------------------------------------------------------------------
// S353
// p(z) = z^{353} + z^{69} + 1

// FIXME dummy value
static mp12_t gf_s353_b = {0x7b13545f, 0x4f50ae31, 0xd57a55aa, 0x72822f6c, 0xa9a197b2, 0xd6ac27c8, 0x4761fa99, 0xf1f3dd67, 0x7fd6422e, 0x3b7b476b, 0x5c4b9a75, 0x0};

#define gf_s353_add mp12_eor

void gf_s353_mod (mp12_t r, unsigned* a);

void gf_s353_square_t (mp12_t r, const mp12_t a);
void gf_s353_square_v (mp12_t r, const mp12_t a);
#define gf_s353_square gf_s353_square_v

void gf_s353_mul_kld1 (mp12_t r, const mp12_t a, const mp12_t b);
void gf_s353_mul_kld  (mp12_t r, const mp12_t a, const mp12_t b);
void gf_s353_mul_kldu (mp12_t r, const mp12_t a, const mp12_t b);
void gf_s353_mul_ld   (mp12_t r, const mp12_t a, const mp12_t b);
void gf_s353_mul_ldu  (mp12_t r, const mp12_t a, const mp12_t b);
void gf_s353_mul_vld  (mp12_t r, const mp12_t a, const mp12_t b);
void gf_s353_mul_v    (mp12_t r, const mp12_t a, const mp12_t b);
void gf_s353_muld_vld (unsigned* r, const mp12_t a, const mp12_t b);
void gf_s353_muld_v   (unsigned* r, const mp12_t a, const mp12_t b);
#define gf_s353_mul   gf_s353_mul_v
#define gf_s353_muld   gf_s353_muld_v
#define gf_s353_mul_b(r, a) gf_s353_mul(r, a, gf_s353_b)

void gf_s353_inv_eeb (mp12_t r, const mp12_t a);
void gf_s353_inv_ee  (mp12_t r, const mp12_t a);
void gf_s353_inv_vee (mp12_t r, const mp12_t a);
#define gf_s353_inv  gf_s353_inv_vee

void gf_s353_sqrt (mp12_t r, const mp12_t a);

void gf_s353_solve_b    (mp12_t r, const mp12_t a);
void gf_s353_solve_rm2  (mp12_t r, const mp12_t a);
void gf_s353_solve_4m2  (mp12_t r, const mp12_t a);
void gf_s353_solve_vb   (mp12_t r, const mp12_t a);
void gf_s353_solve_vrm2 (mp12_t r, const mp12_t a);
void gf_s353_solve_v4m2 (mp12_t r, const mp12_t a);
#define gf_s353_solve   gf_s353_solve_v4m2

int gf_s353_degree (const mp12_t a, int size);

// Trace function of a field element: Tr(z^i) = 1 only for i \in {0} -> Tr(a) = a_0
#define gf_s353_trace(a) (a[0] & 1)

#ifdef FB_POLYN
#if FB_POLYN == 163
#define FNAME b163
#elif FB_POLYN == 283
#define FNAME b283
#elif FB_POLYN == 571
#define FNAME b571
#elif FB_POLYN == 353
#define FNAME s353
#elif FB_POLYN == 251
#define FNAME r251
#endif

#define JOIN(x, y, z) JOIN_AGAIN(x, y, z)
#define JOIN_AGAIN(x, y, z) x ## y ## z
#define gf_mul_ldu JOIN(gf_, FNAME, _mul_ldu)
#define gf_mul_vld JOIN(gf_, FNAME, _mul_vld)
#define gf_square_t JOIN(gf_, FNAME, _square_t)
#define gf_inv_ee JOIN(gf_, FNAME, _inv_ee)
#define gf_inv_vee JOIN(gf_, FNAME, _inv_vee)
#define gf_mul_b JOIN(gf_, FNAME, _mul_b)
#endif


#endif /* ECC_GF_H_ */

