/*
 * RELIC is an Efficient LIbrary for Cryptography
 * Copyright (C) 2007-2012 RELIC Authors
 *
 * This file is part of RELIC. RELIC is legal property of its developers,
 * whose names are not listed here. Please refer to the COPYRIGHT file
 * for contact information.
 *
 * RELIC is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * RELIC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with RELIC. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file
 *
 * Implementation of ARM-dependent routines.
 *
 * @version $Id$
 * @ingroup arch
 */

#include "relic_util.h"

/*============================================================================*/
/* Public definitions                                                         */
/*============================================================================*/

void arch_init(void) {
#if TIMER == CYCLE
	int value = 1; //enable all counters
	value |= 4; //reset cycle counter to zero.
	value |= 8; //enable "by 64" divider for CCNT.
	value |= 16;
	value |= 32;
	unsigned int tmp;

	__asm__ volatile ("MRC p15, 0, %0, c9, c14, 0\t\n" : "=r"(tmp));
	if (tmp & 1) {
		puts("PMUSERENR ENABLED");
	} else {
		puts("PMUSERENR DISABLED");
	}

	//program the performance-counter control-register
	__asm__ volatile ("MCR p15, 0, %0, c9, c12, 0\t\n" :: "r"(value));
	//enable all counters
	__asm__ volatile ("MCR p15, 0, %0, c9, c12, 1\t\n" :: "r"(0x80000000));
	//clear overflows
	__asm__ volatile ("MCR p15, 0, %0, c9, c12, 3\t\n" :: "r"(0x80000000));
#endif
}

void arch_clean(void) {
}

unsigned long long arch_cycles(void) {
#if TIMER == CYCLE
	unsigned int value;
	// Read CCNT Register
	__asm__ volatile ("MRC p15, 0, %0, c9, c13, 0\t\n": "=r"(value));
	return value;
#else
	return 0;
#endif
}
