#include "relic_fb_low.h"
#include "relic_fb_rdc_low_283.s"

#include "util.s"
#include "low.s"

.arm
.align
.globl fb_muln_low_asm
.globl fb_mulm_low_asm
.globl asmbench
.globl mul64bench

asmtest:
    vdup.32 q2, r2
    vldm r1, {q1}
    vshl.U64 q0, q1, q2
    vstm r0, {q0}
    bx lr

dummy:
    push {lr}
    pop {pc}

asmbench:
    .rept 100
    vmov d0, r0, r1
    veor d0, d1
    .endr
    bx lr

# Input
#  ad: A (64bit)
#  bd: B (64bit)
#  k16: $0x000000000000ffff
#  k32: $0x00000000ffffffff
#  k48: $0x0000ffffffffffff

# Output
#  rq (rh|rl): C = A x B (128bit)

# Temporary
#  t0q (t0h|t0l): temporary (128bit)
#  t1q (t1h|t1l): temporary (128bit)
#  t2q (t2h|t2l): temporary (128bit)
#  t3q (t3h|t3l): temporary (128bit)

.macro MUL64k3t4 rq rl rh ad bd k16 k32 k48 t0q t0l t0h t1q t1l t1h t2q t2l t2h t3q t3l t3h

    # a[76543210] * b[07654321]             # a[76543210] * b[10765432]             # a[76543210] * b[21076543]
    # a[07654321] * b[76543210]             # a[10765432] * b[76543210]             # a[21076543] * b[76543210]             # a[76543210] * b[32107654]             # a[76543210] * b[76543210]
    @A1
    vext.8  \t0l, \ad, \ad, $1
    @F = A1*B
    vmull.p8 \t0q, \t0l, \bd
    @B1
    vext.8  \rl, \bd, \bd, $1
    @E = A*B1 (7)
    vmull.p8 \rq, \ad, \rl
                                            @A2
                                            vext.8  \t1l, \ad, \ad, $2
                                            @H = A2*B
                                            vmull.p8 \t1q, \t1l, \bd
                                            @B2
                                            vext.8  \t3l, \bd, \bd, $2
                                            @G = A*B2
                                            vmull.p8 \t3q, \ad, \t3l
                                                                                    @A3
                                                                                    vext.8  \t2l, \ad, \ad, $3
                                                                                    @J = A3*B
                                                                                    vmull.p8 \t2q, \t2l, \bd
    @L = E + F
    veor    \t0q, \t0q, \rq
                                                                                    @B3
                                                                                    vext.8  \rl, \bd, \bd, $3
                                                                                    @I = A*B3
                                                                                    vmull.p8 \rq, \ad, \rl
                                            @M = G + H
                                            veor    \t1q, \t1q, \t3q
                                                                                                                            @B4
                                                                                                                            vext.8  \t3l, \bd, \bd, $4
                                                                                                                            @K = A*B4
                                                                                                                            vmull.p8 \t3q, \ad, \t3l
    @t0 = (L) (P0 + P1) << 8
    veor    \t0l, \t0l, \t0h
    vand    \t0h, \t0h, \k48
                                            @t1 = (M) (P2 + P3) << 16
                                            veor    \t1l, \t1l, \t1h
                                            vand    \t1h, \t1h, \k32
                                                                                    @N = I + J
                                                                                    veor    \t2q, \t2q, \rq
    veor    \t0l, \t0l, \t0h
                                            veor    \t1l, \t1l, \t1h
                                                                                    @t2 = (N) (P4 + P5) << 24
                                                                                    veor    \t2l, \t2l, \t2h
                                                                                    vand    \t2h, \t2h, \k16
                                                                                                                            @t3 = (K) (P6 + P7) << 32
                                                                                                                            veor    \t3l, \t3l, \t3h
                                                                                                                            vmov.i64 \t3h, $0
    vext.8  \t0q, \t0q, \t0q, $15
                                                                                    veor    \t2l, \t2l, \t2h
                                            vext.8  \t1q, \t1q, \t1q, $14
                                                                                                                                                                    vmull.p8 \rq, \ad, \bd
                                                                                    vext.8  \t2q, \t2q, \t2q, $13
                                                                                                                            vext.8  \t3q, \t3q, \t3q, $12
    veor    \t0q, \t0q, \t1q
    veor    \t2q, \t2q, \t3q
    veor    \rq, \rq, \t0q
    veor    \rq, \rq, \t2q

.endm

.macro mul32m rq, rl, rh, ad bd k16 k32 k48 t0q t0l t0h t1q t1l t1h t2q t2l t2h t3q t3l t3h
   @A1
   vext.8  \t0l, \ad, \ad, #1
   @F = A1*B0
   vmull.p8    \t0q, \t0l, \bd
   @B1
   vext.8  \t1l, \bd, \bd, #1
   @E = A0*B1
   vmull.p8    \t1q, \ad, \t1l
   @A2
   vext.8  \t2l, \ad, \ad, #2
   @H = A2*B
   vmull.p8    \t2q, \t2l, \bd
   @B2
   vext.8  \rl, \bd, \bd, #2
   @G = A*B2
   vmull.p8    \rq, \ad, \rl
   @A3
   vext.8  \t3l, \ad, \ad, #3
   @J = A3*B
   vmull.p8    \t3q, \t3l, \bd
   @LL = EL + FL
   veor    \t0l, \t0l, \t1l
   @B3
   vext.8  \t1l, \bd, \bd, #3
   @I = A*B3
   vmull.p8    \t1q, \ad, \t1l
   @ML = GL + HL
   veor    \t2l, \t2l, \rl
   @D = A*B
   vmull.p8    \rq, \ad, \bd
   @(P0+P1)<<8
   vext.8  \t0l, \t0l, \t0l, #7
   @(P2+P3)<<16
   vext.8  \t2l, \t2l, \t2l, #6
   @NL = IL + JL
   veor    \t3l, \t3l, \t1l
   @(P0+P1)<<8 + (P2+P3)<<16
   veor    \t0l, \t0l, \t2l
   @(P4+P5)<<24
   vext.8  \t3l, \t3l, \t3l, #5
   @D + (P0+P1)<<8 + (P2+P3)<<16
   veor    \rl, \rl, \t0l
   @D + (P0+P1)<<8 + (P2+P3)<<16 + (P4+P5)<<24
   veor    \rl, \rl, \t3l
.endm


.macro mul64m res, inlow, inhigh
    @res: the result qword register
    @clobbers all except q4-q8,q11

    @B1
    vext.8 d2, \inhigh, \inhigh, #1
    @A1
    vext.8 d3, \inlow, \inlow, #1
    @B2
    vext.8 d4, \inhigh, \inhigh, #2
    @A2
    vext.8 d5, \inlow, \inlow, #2
    @B3
    vext.8 d6, \inhigh, \inhigh, #3
    @A3
    vext.8 d7, \inlow, \inlow, #3
    @B4
    vext.8 d31, \inhigh, \inhigh, #4

    @E = A*B1 (7)
    vmull.p8    q9, \inlow, d2
    @F = A1*B
    vmull.p8    q12, \inhigh, d3
    @G = A*B2
    vmull.p8    q1, \inlow, d4
    @H = A2*B
    vmull.p8    q13, \inhigh, d5
    @I = A*B3
    vmull.p8    q2, \inlow, d6
    @J = A3*B
    vmull.p8    q10, \inhigh, d7
    @K = A*B4
    vmull.p8    q3, \inlow, d31

    @L = E + F (8)
    veor q9, q9, q12

    @M = G + H
    veor q1, q1, q13

    @q9 = (L) (P0 + P1) << 8 (8)
    veor d18, d18, d19
    @N = I + J (2)
    veor q2, q2, q10
    vand d19, d19, d28

    @q1 = (M) (P2 + P3) << 16 (9)
    veor d2, d2, d3
    vand d3, d3, d29

    @q2 = (N) (P4 + P5) << 24 (8)
    veor d4, d4, d5
    vand d5, d5, d30

    @q9... (L)
    veor d18, d18, d19

    @q1.... (M)
    veor d2, d2, d3

    @q2... (N)
    veor d4, d4, d5

    @q3 = (K) (P6 + P7) << 32 (6)
    veor d6, d6, d7
    veor d7, d7, d7

    @D = A*B
    vmull.p8    q10, \inlow, \inhigh

    @q9... (L)
    vext.8 q9, q9, q9, #15

    @q1.... (M)
    vext.8 q1, q1, q1, #14

    @q2... (N)
    vext.8 q2, q2, q2, #13

    @q3... (K)
    vext.8 q3, q3, q3, #12

    @C (9)
    veor q10, q10, q9
    veor q0, q1, q2
    veor q10, q10, q3
    veor \res, q0, q10
.endm

.macro mul283m2
    @bic ip, sp, #15
    @vstmdb  ip!, {d8-d15}

    vmov.i64 d10, #0
    vmov.i64 d11, #0
    vldm r1!, {d16-d19}
    vldm r2!, {d20-d23}
    vld1.32 {d10[0]}, [r1]
    vld1.32 {d11[0]}, [r2]

    vmov.i64 d14, #0x0000FFFFFFFFFFFF
    vmov.i64 d13, #0x00000000FFFFFFFF
    vmov.i64 d12, #0x000000000000FFFF

    //v0
    MUL64k3t4 q0, d0, d1, d16, d20, d12, d13, d14, q12, d24, d25, q13, d26, d27, q14, d28, d29, q15, d30, d31

    //v1
    MUL64k3t4 q4, d8, d9, d17, d21, d12, d13, d14, q12, d24, d25, q13, d26, d27, q14, d28, d29, q15, d30, d31

    vmov q1, q0
    vmov q3, q0
    veor q2, q0, q4

    veor d2, d1
    veor d6, d1
    veor d1, d0

    veor d2, d9
    veor d5, d0
    veor d1, d8

    //v3
    mul32m q4, d8, d9, d10, d11, d12, d13, d14, q12, d24, d25, q13, d26, d27, q14, d28, d29, q15, d30, d31

    veor d2, d8
    veor d3, d8
    veor d4, d8
    veor d6, d8
    veor d7, d8
    vmov d15, d8

    //v2
    MUL64k3t4 q4, d8, d9, d19, d23, d12, d13, d14, q12, d24, d25, q13, d26, d27, q14, d28, d29, q15, d30, d31

    veor d17, d16
    veor d21, d20

    veor q2, q4
    veor d7, d8
    veor d15, d9

    //v5
    MUL64k3t4 q4, d8, d9, d17, d21, d12, d13, d14, q12, d24, d25, q13, d26, d27, q14, d28, d29, q15, d30, d31

    veor d19, d10
    veor d23, d11

    veor d1, d8
    veor q1, q4
    veor q2, q4
    veor d2, d9
    veor d6, d9
    veor d5, d8

    //v6
    MUL64k3t4 q4, d8, d9, d19, d23, d12, d13, d14, q12, d24, d25, q13, d26, d27, q14, d28, d29, q15, d30, d31

    veor d17, d18
    veor d21, d22
    veor d17, d10
    veor d21, d11

    veor d4, d9
    veor d3, d8
    veor q2, q4
    veor q3, q4
    veor d15, d9
    veor d7, d8

    //v10
    MUL64k3t4 q4, d8, d9, d17, d21, d12, d13, d14, q12, d24, d25, q13, d26, d27, q14, d28, d29, q15, d30, d31

    veor d19, d18
    veor d23, d22
    veor d19, d16
    veor d23, d20

    veor q1, q4
    veor d5, d8
    veor d6, d9

    //v11
    MUL64k3t4 q4, d8, d9, d19, d23, d12, d13, d14, q12, d24, d25, q13, d26, d27, q14, d28, d29, q15, d30, d31

    veor d17, d16
    veor d21, d20

    veor d3, d8
    veor d4, d9
    veor d6, d8
    veor d7, d9

    //v7
    MUL64k3t4 q4, d8, d9, d17, d21, d12, d13, d14, q12, d24, d25, q13, d26, d27, q14, d28, d29, q15, d30, d31

    veor d19, d10
    veor d23, d11

    veor q1, q4
    veor q2, q4

    //v8
    MUL64k3t4 q4, d8, d9, d19, d23, d12, d13, d14, q12, d24, d25, q13, d26, d27, q14, d28, d29, q15, d30, d31

    veor d16, d10
    veor d20, d11

    veor q2, q4
    veor q3, q4

    //v4
    MUL64k3t4 q4, d8, d9, d16, d20, d12, d13, d14, q12, d24, d25, q13, d26, d27, q14, d28, d29, q15, d30, d31

    veor d19, d17
    veor d23, d21

    veor q3, q4
    veor q1, q4
    veor d4, d9
    veor d3, d8
    veor d5, d8
    veor d6, d9

    //v9
    MUL64k3t4 q4, d8, d9, d19, d23, d12, d13, d14, q12, d24, d25, q13, d26, d27, q14, d28, d29, q15, d30, d31

    veor d18, d19
    veor d22, d23

    veor d3, d8
    veor d4, d9
    veor d5, d8
    veor d6, d9

    //v12
    MUL64k3t4 q4, d8, d9, d18, d22, d12, d13, d14, q12, d24, d25, q13, d26, d27, q14, d28, d29, q15, d30, d31

    veor q2, q4
    veor d3, d8
    veor d4, d9
    veor d5, d8
    veor d6, d9

    vmov d8, d15

.endm


.macro mul283m
    sub sp, sp, #7*16

    vmov.i64 d28, #0x0000FFFFFFFFFFFF
    vmov.i64 d29, #0x00000000FFFFFFFF
    vmov.i64 d30, #0x000000000000FFFF

    veor d12, d12
    veor d17, d17
    vldm r1!, {d8, d9, d10, d11}
    vldm r2!, {d13, d14, d15, d16}
    vld1.32 {d12[0]}, [r1]
    vld1.32 {d17[0]}, [r2]

    vswp d9, d10
    vswp d15, d16
    vswp d9, d13
    vswp d12, d16
    vswp d11, d14
    vswp d12, d13

    //q4 = a0:b0
    //q5 = a1:b1
    //q6 = a2:b2
    //q7 = a3:b3
    //q8 = a4:b4

    mov r3, sp

    @mul64m q0, d8, d9
    MUL64k3t4 q0, d0, d1, d8, d9, d30, d29, d28, q1, d2, d3, q2, d4, d5, q3, d6, d7, q9, d18, d19
    //m[0]
    vstm r3!, {q0}

    @mul64m q0, d10, d11
    MUL64k3t4 q0, d0, d1, d10, d11, d30, d29, d28, q1, d2, d3, q2, d4, d5, q3, d6, d7, q9, d18, d19
    //m[1]
    vstm r3!, {q0}

    @mul64m q0, d12, d13
    MUL64k3t4 q0, d0, d1, d12, d13, d30, d29, d28, q1, d2, d3, q2, d4, d5, q3, d6, d7, q9, d18, d19
    //m[2]
    vstm r3!, {q0}

    @mul64m q0, d14, d15
    MUL64k3t4 q0, d0, d1, d14, d15, d30, d29, d28, q1, d2, d3, q2, d4, d5, q3, d6, d7, q9, d18, d19
    //m[3]
    vstm r3!, {q0}

    @mul64m q1, d16, d17
    @MUL64k3t4 q1, d2, d3, d16, d17, d30, d29, d28, q0, d0, d1, q2, d4, d5, q3, d6, d7, q9, d18, d19
    mul32m q1, d2, d3, d16, d17, d30, d29, d28, q0, d0, d1, q2, d4, d5, q3, d6, d7, q9, d18, d19
    veor d3, d3
    //m[4]
    veor q0, q4, q5
    vstm r3!, {q1}

    //t0 = a0+a1:b0+b1
    @mul64m q1, d0, d1
    MUL64k3t4 q1, d2, d3, d0, d1, d30, d29, d28, q10, d20, d21, q2, d4, d5, q3, d6, d7, q9, d18, d19
    //m[5]
    veor q0, q4, q6
    vstm r3!, {q1}

    //t1 = a0+a2:b0+b2
    @mul64m q1, d0, d1
    MUL64k3t4 q1, d2, d3, d0, d1, d30, d29, d28, q10, d20, d21, q2, d4, d5, q3, d6, d7, q9, d18, d19
    //m[6]
    veor q0, q6, q8
    vstm r3!, {q1}
    veor q13, q7, q8
    //q4 = a0 + a3
    veor q4, q4, q7

    //t2 = a2+a4:b2+b4
    @mul64m q1, d0, d1
    MUL64k3t4 q7, d14, d15, d0, d1, d30, d29, d28, q10, d20, d21, q2, d4, d5, q3, d6, d7, q9, d18, d19
    //m[7]
    //q5 = a1 + a4
    veor q5, q5, q8

    //t3 = a3+a4:b3+b4
    @mul64m q1, d26, d27
    MUL64k3t4 q8, d16, d17, d26, d27, d30, d29, d28, q10, d20, d21, q2, d4, d5, q3, d6, d7, q9, d18, d19
    //m[8]

    //t1 = (a0+a2)+a3:(b0+b2)+b3
    veor q0, q4, q6
    @mul64m q1, d0, d1
    MUL64k3t4 q9, d18, d19, d0, d1, d30, d29, d28, q10, d20, d21, q2, d4, d5, q3, d6, d7, q1, d2, d3
    //m[9]
    veor q0, q5, q6

    //t2 = (a2+a4)+a1:(b2+b4)+b1
    @mul64m q1, d0, d1
    MUL64k3t4 q10, d20, d21, d0, d1, d30, d29, d28, q13, d26, d27, q2, d4, d5, q3, d6, d7, q1, d2, d3
    veor q4, q4, q5
    //m[10]

    //t3 = (a3+a4)+(a0+a1):(b3+b4)+(b0+b1)
    @mul64m q11, d8, d9
    MUL64k3t4 q11, d22, d23, d8, d9, d30, d29, d28, q13, d26, d27, q2, d4, d5, q3, d6, d7, q1, d2, d3
    //m[11]

    //t0 = (a3+a4)+(a0+a1)+a2:(b3+b4)+(b0+b1)+b2
    veor q0, q4, q6
    @mul64m q12, d0, d1
    MUL64k3t4 q12, d24, d25, d0, d1, d30, d29, d28, q13, d26, d27, q2, d4, d5, q3, d6, d7, q1, d2, d3
    //m[12]

    //c0(q0) = m[0];
    //c8(q4) = m[4]
    vldm sp!, {d0-d13}

    //c1(q1) = c0 + m1
    veor q1, q0, q1
    //c4(q9) = m9 + m12
    veor q9, q9, q12
    //c2(q6) = c1 + m6
    veor q6, q1, q6
    //c7(q3) = m3 + c8
    veor q3, q3, q4
    //c2 = c2 + m2
    veor q6, q6, q2
    //c6(q7) = c7 + m7
    veor q7, q3, q7
    //c5(q13) = c2 + m10
    veor q13, q6, q10
    //c6 = c6 + m2
    veor q7, q7, q2
    //c5 = c5 + m11
    veor q13, q13, q11
    //c3(q2) = c4 + c6
    veor q2, q9, q7
    //c7 = c7 + m8
    veor q3, q3, q8
    //c1 = c1 + m5
    veor q1, q1, q5
    //c5 = c5 + m12
    veor q13, q13, q12
    //c4 = c4 + c1
    veor q9, q9, q1
    //c3 = c3 + c0
    veor q2, q2, q0
    //c4 = c4 + c7
    veor q9, q9, q3
    //c5 = c5 + c8
    veor q13, q13, q4
    //c3 = c3 + m11
    veor q2, q2, q11
    //c4 = c4 + m10
    veor q9, q9, q10

    //c0  c1  c2  c3  c4  c5  c6  c7  c8
    //q0  q1  q6  q2  q9  q13 q7  q3  q4
    //q0      q1      q2      q3      q4

    veor d1, d1, d2
    veor d8, d8, d7
    veor d7, d15, d6
    veor d6, d14, d27
    veor d2, d12, d3
    veor d3, d13, d4
    veor d4, d18, d5
    veor d5, d19, d26
.endm

fb_muln_low_asm:
    mul283m2
    vstm r0!, {d0-d8}
    bx lr

fb_mulm_low_asm:
    bic ip, sp, #15
    vstmdb  ip!, {d8-d15}
    mul283m2
    rdc283m q0, d0, d1, q1, d2, d3, d4, q0, d0, d1, q1, d2, d3, q2, d4, d5, q3, d6, d7, d8
    vstm r0, {d0-d4}
    vldmia  ip!, {d8-d15}
    bx lr

mul64full:
    veor q15, q15, q15
    vorn d30, d30, d30
    vext.8 d28, d30, d31, #2
    vext.8 d29, d30, d31, #4
    vext.8 d30, d30, d31, #6

mul64:
    MUL64k3t4 q0, d0, d1, d8, d9, d30, d29, d28, q1, d2, d3, q2, d4, d5, q3, d6, d7, q9, d18, d19
    bx lr

mul64bench:
    @veor q15, q15, q15
    @vorn d30, d30, d30
    @vext.8 d28, d30, d31, #2
    @vext.8 d29, d30, d31, #4
    @vext.8 d30, d30, d31, #6
    .rept 50
    MUL64k3t4 q0, d0, d1, d8, d9, d30, d29, d28, q1, d2, d3, q2, d4, d5, q3, d6, d7, q9, d18, d19
    MUL64k3t4 q10, d20, d21, d8, d9, d30, d29, d28, q1, d2, d3, q2, d4, d5, q3, d6, d7, q9, d18, d19
    .endr
    bx lr
