/*
 * RELIC is an Efficient LIbrary for Cryptography
 * Copyright (C) 2007-2012 RELIC Authors
 *
 * This file is part of RELIC. RELIC is legal property of its developers,
 * whose names are not listed here. Please refer to the COPYRIGHT file
 * for contact information.
 *
 * RELIC is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * RELIC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with RELIC. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file
 *
 * Implementation of the low-level binary field bit shifting functions.
 *
 * @version $Id$
 * @ingroup fb
 */

#include "relic_fb.h"
#include "relic_util.h"
#include "relic_fb_low.h"
#include "ecc_gf.h"

/*============================================================================*/
/* Public definitions                                                         */
/*============================================================================*/

dig_t fb_lsh1_low(dig_t *c, dig_t *a) {
#if FB_POLYN == 163
	return mp6_lsl_1(c, a);
#elif FB_POLYN == 233
	return mp8_lsl_1(c, a);
#elif FB_POLYN == 283
	return mp9_lsl_1(c, a);
#elif FB_POLYN == 409
	return mp13_lsl_1(c, a);
#elif FB_POLYN == 571
	return mp18_lsl_1(c, a);
#elif FB_POLYN == 251
	return mp8_lsl_1(c, a);
#elif FB_POLYN == 271
	return mp9_lsl_1(c, a);
#elif FB_POLYN == 353
    return mp12_lsl_1(c, a);
#endif
}

dig_t fb_lshb_low(dig_t *c, dig_t *a, int bits) {
#if FB_POLYN == 163
	return mp6_lsl_32(c, a, bits);
#elif FB_POLYN == 233
	return mp8_lsl_32(c, a, bits);
#elif FB_POLYN == 283
	return mp9_lsl_32(c, a, bits);
#elif FB_POLYN == 409
	return mp13_lsl_32(c, a, bits);
#elif FB_POLYN == 571
	return mp18_lsl_32(c, a, bits);
#elif FB_POLYN == 251
	return mp8_lsl_32(c, a, bits);
#elif FB_POLYN == 271
	return mp9_lsl_32(c, a, bits);
#elif FB_POLYN == 353
    return mp12_lsl_32(c, a, bits);
#endif
}

void fb_lshd_low(dig_t *c, dig_t *a, int digits) {
	dig_t *top, *bot;
	int i;

	top = c + FB_DIGS - 1;
	bot = a + FB_DIGS - 1 - digits;

	for (i = 0; i < FB_DIGS - digits; i++, top--, bot--) {
		*top = *bot;
	}
	for (i = 0; i < digits; i++, c++) {
		*c = 0;
	}
}

dig_t fb_rsh1_low(dig_t *c, dig_t *a) {
#if FB_POLYN == 163
	return mp6_lsr_1(c, a);
#elif FB_POLYN == 233
	return mp8_lsr_1(c, a);
#elif FB_POLYN == 283
	return mp9_lsr_1(c, a);
#elif FB_POLYN == 409
	return mp13_lsr_1(c, a);
#elif FB_POLYN == 571
	return mp18_lsr_1(c, a);
#elif FB_POLYN == 251
	return mp8_lsr_1(c, a);
#elif FB_POLYN == 271
	return mp9_lsr_1(c, a);
#elif FB_POLYN == 353
    return mp12_lsr_1(c, a);
#endif
}

dig_t fb_rshb_low(dig_t *c, dig_t *a, int bits) {
#if FB_POLYN == 163
	return mp6_lsr_32(c, a, bits);
#elif FB_POLYN == 233
	return mp8_lsr_32(c, a, bits);
#elif FB_POLYN == 283
	return mp9_lsr_32(c, a, bits);
#elif FB_POLYN == 409
	return mp13_lsr_32(c, a, bits);
#elif FB_POLYN == 571
	return mp18_lsr_32(c, a, bits);
#elif FB_POLYN == 251
	return mp8_lsr_32(c, a, bits);
#elif FB_POLYN == 271
	return mp9_lsr_32(c, a, bits);
#elif FB_POLYN == 353
    return mp12_lsr_32(c, a, bits);
#endif
}

void fb_rshd_low(dig_t *c, dig_t *a, int digits) {
	dig_t *top, *bot;
	int i;

	top = a + digits;
	bot = c;

	for (i = 0; i < FB_DIGS - digits; i++, top++, bot++) {
		*bot = *top;
	}
	for (; i < FB_DIGS; i++, bot++) {
		*bot = 0;
	}
}

dig_t fb_lshadd_low(dig_t *c, dig_t *a, int bits, int size) {
	int i, j;
	dig_t b1, b2;

	j = FB_DIGIT - bits;
	b1 = a[0];
	c[0] ^= (b1 << bits);
	if (size == FB_DIGS) {
		for (i = 1; i < FB_DIGS; i++) {
			b2 = a[i];
			c[i] ^= ((b2 << bits) | (b1 >> j));
			b1 = b2;
		}
	} else {
		for (i = 1; i < size; i++) {
			b2 = a[i];
			c[i] ^= ((b2 << bits) | (b1 >> j));
			b1 = b2;
		}
	}
	return (b1 >> j);
}
